"""Apply package name normalization."""
from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.requirements import Requirement
from tomlkit.api import string as toml_string

from .util import sorted_array

if TYPE_CHECKING:
    from tomlkit.items import Array, String


def normalize_req(req: str) -> str:
    """
    Normalize a python requirement.

    :param req: the raw requirement
    :return: normalized name
    """
    parsed = Requirement(req)
    for spec in parsed.specifier:
        if spec.operator in (">=", "=="):
            version = spec.version
            while version.endswith(".0"):
                version = version[:-2]
            spec._spec = (spec._spec[0], version)  # noqa: SLF001
    return str(parsed)


def _best_effort_string_repr(req: str) -> String:
    # Convert requirement to a TOML string, choosing the most appropriate representation (basic or literal).
    # This function will attempt to use literal strings to avoid escaping double-quotes ("), if the requirement value
    # allows it, e.g. it does not contain other reserved characters such as single quotes (') or newlines (\\n).
    try:
        return toml_string(req, literal=('"' in req))
    except ValueError:
        return toml_string(req)


def normalize_pep508_array(requires_array: Array | None, indent: int, *, keep_full_version: bool) -> None:
    """
    Normalize a TOML array via PEP-508.

    :param requires_array: the input array
    :param indent: indentation level
    :param keep_full_version: whether to preserve, and therefore not normalize, requirements
    """
    if requires_array is None:
        return
    # first normalize values
    for at in range(len(requires_array)):
        initial_requirement_string = str(requires_array[at])
        req_string = initial_requirement_string if keep_full_version else normalize_req(initial_requirement_string)
        normalized = _best_effort_string_repr(req=req_string)
        requires_array[at] = normalized
    # then sort
    sorted_array(requires_array, indent, key=lambda e: Requirement(e.text).name.lower())


__all__ = [
    "normalize_req",
    "normalize_pep508_array",
]



from __future__ import annotations

from typing import Callable, Union

from tomlkit.toml_document import TOMLDocument

from pyproject_fmt.formatter.config import Config

Fmt = Callable[[Callable[[TOMLDocument, Config], None], Union[str, Config], str], None]

__all__ = [
    "Fmt",
]



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from pyproject_fmt.formatter.build_system import fmt_build_system
from pyproject_fmt.formatter.config import Config

if TYPE_CHECKING:
    from tests import Fmt


def test_build_system_missing(fmt: Fmt) -> None:
    fmt(fmt_build_system, "", "\n")


def test_backend_path(fmt: Fmt) -> None:
    start = "[build-system]\nbackend-path = ['A-B', 'AA']"
    expected = "[build-system]\nbackend-path = [\n  'A-B',\n  'AA',\n]\n"
    fmt(fmt_build_system, start, expected)


def test_requires(fmt: Fmt) -> None:
    txt = """
        [build-system]
        requires = [
        # start
        # two
        "A-B", "A>1",     # c # d
        # follow-up comment
        "C",
        # magic
        "D"  ,   # a
        "E",  # b
        "F" # comment
            # ok
        ]
        """

    expected = """
        [build-system]
        requires = [
          # start
          # two
          "A>1", # c # d follow-up comment
          "A-B",
          "C", # magic
          "D", # a
          "E", # b
          "F", # comment ok
        ]
        """

    fmt(fmt_build_system, txt, expected)


def test_build_backend_order(fmt: Fmt) -> None:
    txt = """
    [build-system]
    backend-path = ['A']
    requires = ["A"]
    build-backend = "hatchling.build"
    """

    expected = """
    [build-system]
    build-backend = "hatchling.build"
    requires = [
      "A",
    ]
    backend-path = [
      'A',
    ]
    """

    fmt(fmt_build_system, txt, expected)


@pytest.mark.parametrize("indent", [0, 2, 4])
def test_indent(fmt: Fmt, indent: int) -> None:
    txt = """
    [build-system]
    requires = [
    "A",
    "B",
    ]
    backend-path = [
    "C",
    ]
    """
    expected = f"""
    [build-system]
    requires = [
    {" " * indent}"A",
    {" " * indent}"B",
    ]
    backend-path = [
    {" " * indent}"C",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=indent)
    fmt(fmt_build_system, config, expected)


def test_keep_full_version_on(fmt: Fmt) -> None:
    txt = """
    [build-system]
    requires = [
      "A==1.0.0",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=2, keep_full_version=True)
    fmt(fmt_build_system, config, txt)


def test_keep_full_version_off(fmt: Fmt) -> None:
    txt = """
    [build-system]
    requires = [
      "A==1.0.0",
    ]
    """
    expected = """
    [build-system]
    requires = [
      "A==1",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=2, keep_full_version=False)
    fmt(fmt_build_system, config, expected)



from __future__ import annotations

from pathlib import Path
from subprocess import CalledProcessError
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from pyproject_fmt.formatter.config import Config
from pyproject_fmt.formatter.project import fmt_project

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tests import Fmt


@pytest.mark.parametrize(
    "value",
    [
        "[project]\nname='a-b'",
        "[project]\nname='A_B'",
        "[project]\nname='a.-..-__B'",
    ],
)
def test_project_name(fmt: Fmt, value: str) -> None:
    fmt(fmt_project, value, '[project]\nname="a-b"\n')


def test_project_classifiers(fmt: Fmt) -> None:
    start = """
    [project]
    classifiers = [
      "Operating System :: OS Independent",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3 :: Only",
      "License :: OSI Approved :: MIT License",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.12",
      "Programming Language :: Python :: 3.8",
      "License :: OSI Approved :: MIT License",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.11",
    ]
    """
    expected = """
    [project]
    classifiers = [
      "License :: OSI Approved :: MIT License",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_project_dependencies(fmt: Fmt) -> None:
    start = '[project]\ndependencies=["pytest","pytest-cov",]'
    expected = '[project]\ndependencies=[\n  "pytest",\n  "pytest-cov",\n]\n'
    fmt(fmt_project, start, expected)


def test_project_dependencies_with_double_quotes(fmt: Fmt) -> None:
    start = """
    [project]
    dependencies = [
        'packaging>=20.0;python_version>"3.4"',
        "appdirs"
    ]
    """
    expected = """
    [project]
    dependencies = [
      "appdirs",
      'packaging>=20; python_version > "3.4"',
    ]
    """
    fmt(fmt_project, start, expected)


def test_project_dependencies_with_mixed_quotes(fmt: Fmt) -> None:
    start = """
    [project]
    dependencies = [
        "packaging>=20.0;python_version>\\"3.4\\" and python_version != '3.5'",
        "foobar@ git+https://weird-vcs/w/index.php?param=org'repo ; python_version == '2.7'",
        "appdirs"
    ]
    """
    expected = """
    [project]
    dependencies = [
      "appdirs",
      "foobar@ git+https://weird-vcs/w/index.php?param=org'repo ; python_version == \\"2.7\\"",
      'packaging>=20; python_version > "3.4" and python_version != "3.5"',
    ]
    """
    fmt(fmt_project, start, expected)


def test_project_description(fmt: Fmt) -> None:
    start = '[project]\ndescription=" Magical stuff\t"'
    expected = '[project]\ndescription="Magical stuff"\n'
    fmt(fmt_project, start, expected)


def test_project_scripts(fmt: Fmt) -> None:
    start = """
    [project.scripts]
    c = "d"
    a = "b"
    """
    expected = """
    [project.scripts]
    a = "b"
    c = "d"
    """
    fmt(fmt_project, start, expected)


def test_project_optional_dependencies(fmt: Fmt) -> None:
    start = """
    [project.optional-dependencies]
    test = ["B", "A"]
    docs = [ "C",
    "D"]
    """
    expected = """
    [project.optional-dependencies]
    docs = [
      "C",
      "D",
    ]
    test = [
      "A",
      "B",
    ]
    """
    fmt(fmt_project, start, expected)


def test_entry_points(fmt: Fmt) -> None:
    start = """
    [project.entry-points]
    beta = {C = "c", D = "d"}
    alpha = {B = "b", "A.A" = "a"}
    """
    expected = """
    [project.entry-points]
    alpha = {"A.A" = "a",B = "b"}
    beta = {C = "c",D = "d"}
    """
    fmt(fmt_project, start, expected)


def test_classifier_none(fmt: Fmt) -> None:
    start = """
    [project]
    """
    fmt(fmt_project, start, start)


def test_classifier_lt(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "<3.7"
    """
    expected = """
    [project]
    requires-python = "<3.7"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_gt(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = ">3.6"
    """
    expected = """
    [project]
    requires-python = ">3.6"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_gte(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = ">=3.6"
    """
    expected = """
    [project]
    requires-python = ">=3.6"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.6",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_eq(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python="==3.12"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    expected = """
    [project]
    requires-python="==3.12"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_neq(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "!=3.9"
    """
    expected = """
    [project]
    requires-python = "!=3.9"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_range(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python=">=3.7,<3.13"
    """
    expected = """
    [project]
    requires-python=">=3.7,<3.13"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_range_neq(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "<=3.12,!=3.9,>=3.8"
    """
    expected = """
    [project]
    requires-python = "<=3.12,!=3.9,>=3.8"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_high_range(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "<=3.13,>3.10"
    """
    expected = """
    [project]
    requires-python = "<=3.13,>3.10"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
      "Programming Language :: Python :: 3.13",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_upper_bound(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "<3.8"
    classifiers = [
      "Programming Language :: Python :: 3.5",
      "Programming Language :: Python :: 3.6",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
    ]
    """
    expected = """
    [project]
    requires-python = "<3.8"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_two_upper_bounds(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python = "<3.8,<=3.10"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.5",
      "Programming Language :: Python :: 3.6",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
    ]
    """
    expected = """
    [project]
    requires-python = "<3.8,<=3.10"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.7",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_gt_tox(fmt: Fmt, tmp_path: Path) -> None:
    (tmp_path / "tox.ini").write_text("[tox]\nenv_list = py{311,312}-{magic}")
    start = """
    [project]
    requires-python=">=3.11"
    """
    expected = """
    [project]
    requires-python=">=3.11"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_gt_tox_no_py_ver(fmt: Fmt, tmp_path: Path) -> None:
    (tmp_path / "tox.ini").write_text("[tox]\nenv_list = py-{magic,p12}")
    start = """
    [project]
    requires-python=">=3.11"
    """
    expected = """
    [project]
    requires-python=">=3.11"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_gt_tox_conf_missing(fmt: Fmt) -> None:
    start = """
    [project]
    requires-python=">=3.12"
    """
    expected = """
    [project]
    requires-python=">=3.12"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, expected)


def test_classifier_tox_fails_call(fmt: Fmt, mocker: MockerFixture) -> None:
    mocker.patch(
        "pyproject_fmt.formatter.project.check_output",
        side_effect=CalledProcessError(1, []),
    )

    start = """
    [project]
    requires-python=">=3.12"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, start)


def test_classifier_tox_exe_bad(
    fmt: Fmt,
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("PATH", str(tmp_path))
    tox_bin = tmp_path / "tox"
    tox_bin.write_text("")
    tox_bin.chmod(0o755)

    start = """
    [project]
    requires-python=">=3.12"
    classifiers = [
      "Programming Language :: Python :: 3 :: Only",
      "Programming Language :: Python :: 3.12",
    ]
    """
    fmt(fmt_project, start, start)


@pytest.mark.parametrize("indent", [0, 2, 4])
def test_indent(fmt: Fmt, indent: int) -> None:
    txt = """
    [project]
    keywords = [
      "A",
    ]
    dynamic = [
      "B",
    ]
    classifiers = [
      "C",
    ]
    dependencies = [
      "D",
    ]
    [project.optional-dependencies]
    docs = [
      "E",
    ]
    """
    expected = f"""
    [project]
    keywords = [
    {" " * indent}"A",
    ]
    classifiers = [
    {" " * indent}"C",
    ]
    dynamic = [
    {" " * indent}"B",
    ]
    dependencies = [
    {" " * indent}"D",
    ]
    [project.optional-dependencies]
    docs = [
    {" " * indent}"E",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=indent)
    fmt(fmt_project, config, expected)


def test_keep_full_version_on(fmt: Fmt) -> None:
    txt = """
    [project]
    dependencies = [
      "A==1.0.0",
    ]
    [project.optional-dependencies]
    docs = [
      "B==2.0.0",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=2, keep_full_version=True)
    fmt(fmt_project, config, txt)


def test_keep_full_version_off(fmt: Fmt) -> None:
    txt = """
    [project]
    dependencies = [
      "A==1.0.0",
    ]
    [project.optional-dependencies]
    docs = [
      "B==2.0.0",
    ]
    """
    expected = """
    [project]
    dependencies = [
      "A==1",
    ]
    [project.optional-dependencies]
    docs = [
      "B==2",
    ]
    """
    config = Config(pyproject_toml=Path(), toml=dedent(txt), indent=2, keep_full_version=False)
    fmt(fmt_project, config, expected)



from __future__ import annotations

import os
import sys
from stat import S_IREAD, S_IWRITE
from typing import TYPE_CHECKING

import pytest

from pyproject_fmt._version import __version__
from pyproject_fmt.cli import cli_args

if TYPE_CHECKING:
    from pathlib import Path


def test_cli_version(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit) as context:
        cli_args(["--version"])
    assert context.value.code == 0
    out, err = capsys.readouterr()
    assert out == f"pyproject-fmt ({__version__})\n"


def test_cli_pyproject_toml_ok(tmp_path: Path) -> None:
    path = tmp_path / "tox.ini"
    path.write_text("")
    result = cli_args([str(path)])
    assert result.inputs == [path]


def test_cli_inputs_ok(tmp_path: Path) -> None:
    paths = []
    for filename in ("tox.ini", "tox2.ini", "tox3.ini"):
        path = tmp_path / filename
        path.write_text("")
        paths.append(path)
    result = cli_args([*map(str, paths)])
    assert result.inputs == paths


def test_cli_pyproject_toml_not_exists(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    with pytest.raises(SystemExit) as context:
        cli_args([str(tmp_path / "tox.ini")])
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert "argument inputs: path does not exist" in err


def test_cli_pyproject_toml_not_file(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    path = tmp_path / "temp"
    os.mkfifo(path)
    with pytest.raises(SystemExit) as context:
        cli_args([str(path)])
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert "argument inputs: path is not a file" in err


@pytest.mark.parametrize(("flag", "error"), [(S_IREAD, "write"), (S_IWRITE, "read")])
@pytest.mark.skipif(
    sys.platform == "win32",
    reason="On Windows files cannot be read only, only folders",
)
def test_cli_pyproject_toml_permission_fail(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
    flag: int,
    error: str,
) -> None:
    path = tmp_path / "tox.ini"
    path.write_text("")
    path.chmod(flag)
    try:
        with pytest.raises(SystemExit) as context:
            cli_args([str(path)])
    finally:
        path.chmod(S_IWRITE | S_IREAD)
    assert context.value.code != 0
    out, err = capsys.readouterr()
    assert not out
    assert f"argument inputs: cannot {error} path" in err


def test_pyproject_toml_resolved(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    path = tmp_path / "tox.ini"
    path.write_text("")
    result = cli_args(["tox.ini"])
    assert result.inputs == [path]


def test_pyproject_toml_dir(tmp_path: Path) -> None:
    (tmp_path / "pyproject.toml").write_text("")
    cli_args([str(tmp_path)])



from __future__ import annotations

from textwrap import dedent
from typing import cast

import pytest
from tomlkit import parse
from tomlkit.items import Array

from pyproject_fmt.formatter.pep508 import normalize_pep508_array, normalize_req


@pytest.mark.parametrize(
    ("value", "result"),
    [
        ("a", "a"),
        ('packaging>=20.0;python_version>"3.4"', 'packaging>=20; python_version > "3.4"'),
        (
            "xonsh>=0.9.16;python_version>'3.4' and python_version!='3.9'",
            'xonsh>=0.9.16; python_version > "3.4" and python_version != "3.9"',
        ),
        ("pytest-xdist>=1.31.0", "pytest-xdist>=1.31"),
        ("foo@http://foo.com", "foo@ http://foo.com"),
        (
            "bar [fred,al] @ http://bat.com ;python_version=='2.7'",
            'bar[al,fred]@ http://bat.com ; python_version == "2.7"',
        ),
        (
            "baz [quux, strange];python_version<\"2.7\" and platform_version=='2'",
            'baz[quux,strange]; python_version < "2.7" and platform_version == "2"',
        ),
        ("pytest>=6.0.0", "pytest>=6"),
        ("pytest==6.0.0", "pytest==6"),
        ("pytest~=6.0.0", "pytest~=6.0.0"),
    ],
)
def test_requires_fmt(value: str, result: str) -> None:
    outcome = normalize_req(req=value)
    assert outcome == result


@pytest.mark.parametrize("char", ["!", "=", ">", "<", " ", "\t", "@"])
def test_bad_syntax_requires(char: str) -> None:
    with pytest.raises(ValueError, match=f"[{char}]" if char.strip() else None):
        normalize_req(f"{char};")


@pytest.mark.parametrize("indent", [0, 2, 4])
def test_normalize_pep508_array(indent: int) -> None:
    toml_document_string = """
        requirements = [
            "zzz>=1.1.1",
            "pytest==6.0.0",
        ]
        """
    parsed = parse(toml_document_string)
    dependencies = parsed["requirements"]
    normalize_pep508_array(
        requires_array=cast(Array, dependencies),
        indent=indent,
        keep_full_version=False,
    )
    assert dependencies == ["zzz>=1.1.1", "pytest==6"]
    expected_string = dedent(
        f"""\
        [
        {" " * indent}"pytest==6",
        {" " * indent}"zzz>=1.1.1",
        ]
        """,
    ).strip()
    assert dependencies.as_string() == expected_string


def test_normalize_pep508_array_keep_versions() -> None:
    toml_document_string = """
        requirements = [
            "pytest==6.0.0",
        ]
        """
    parsed = parse(toml_document_string)
    dependencies = parsed["requirements"]
    normalize_pep508_array(
        requires_array=cast(Array, dependencies),
        indent=2,
        keep_full_version=True,
    )
    assert dependencies == ["pytest==6.0.0"]



from __future__ import annotations

from typing import TYPE_CHECKING

from pyproject_fmt.formatter.tools import fmt_tools

if TYPE_CHECKING:
    from tests import Fmt


def test_tools_ordering(fmt: Fmt) -> None:
    content = """
    [tool.poetry]
    name = "a"
    [tool.coverage]
    a = 0
    [tool.poetry.scripts]
    version = "1"
    [tool.scikit-build]
    a = 0
    [tool.pytest]
    a = 0
    [tool.black]
    a = 0
    [tool.isort]
    a = 0
    [tool.flake8]
    [tool.hatch]
    a = 0
    [tool.setuptools_scm]
    a = 0
    [tool.setuptools]
    a.b = 0
    [tool.docformatter]
    c = 0
    [tool.autopep8]
    a = 0
    [tool.distutils]
    a = 0
    [tool.codespell]
    [tool.meson-python]
    a = 0
    [tool.cibuildwheel]
    a = 0
    [tool.pylint]
    [tool.repo-review]
    a = 0
    [tool.bumpversion]
    [tool.check-manifest]
    [tool.check-sdist]
    [tool.check-wheel-contents]
    [tool.doit]
    [tool.flit]
    [tool.jupyter-releaser]
    [tool.maturin]
    [tool.mypy]
    [tool.nbqa]
    [tool.pdm]
    [tool.py-build-cmake]
    [tool.pycln]
    [tool.pydoclint]
    [tool.pyright]
    [tool.pytest-enabler]
    [tool.pytest_env]
    [tool.sphinx-theme-builder]
    [tool.spin]
    [tool.tbump]
    [tool.tomlsort]
    [tool.towncrier]
    [tool.tox]
    [tool.vendoring]
    [tool.whey]
    """
    expected = """
    [tool.poetry]
    name = "a"
    [tool.poetry.scripts]
    version = "1"
    [tool.pdm]

    [tool.setuptools]
    a.b = 0

    [tool.distutils]
    a = 0

    [tool.setuptools_scm]
    a = 0

    [tool.hatch]
    a = 0

    [tool.flit]

    [tool.scikit-build]
    a = 0

    [tool.meson-python]
    a = 0

    [tool.maturin]

    [tool.whey]

    [tool.py-build-cmake]

    [tool.sphinx-theme-builder]

    [tool.cibuildwheel]
    a = 0

    [tool.autopep8]
    a = 0

    [tool.black]
    a = 0

    [tool.isort]
    a = 0

    [tool.flake8]

    [tool.pycln]

    [tool.nbqa]

    [tool.pylint]

    [tool.repo-review]
    a = 0

    [tool.codespell]

    [tool.docformatter]
    c = 0

    [tool.pydoclint]

    [tool.tomlsort]

    [tool.check-manifest]

    [tool.check-sdist]

    [tool.check-wheel-contents]

    [tool.pytest]
    a = 0

    [tool.pytest_env]

    [tool.pytest-enabler]

    [tool.coverage]
    a = 0

    [tool.doit]

    [tool.spin]

    [tool.tox]

    [tool.bumpversion]

    [tool.jupyter-releaser]

    [tool.tbump]

    [tool.towncrier]

    [tool.vendoring]

    [tool.mypy]

    [tool.pyright]
        """
    fmt(fmt_tools, content, expected)


def test_sub_table_newline(fmt: Fmt) -> None:
    content = """
    [tool.mypy]
    a = 0
    [[tool.mypy.overrides]]
    a = 1
    [tool.something-else]
    b = 0
    """
    expected = """
    [tool.mypy]
    a = 0
    [[tool.mypy.overrides]]
    a = 1

    [tool.something-else]
    b = 0
    """
    fmt(fmt_tools, content, expected)


def test_sub_table_no_op(fmt: Fmt) -> None:
    content = """
    [tool.mypy]
    a = 0

    [[tool.mypy.overrides]]
    a = 1

    [tool.something-else]
    b = 0
    """
    fmt(fmt_tools, content, content)



"""CLI interface parser."""
from __future__ import annotations

import os
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    ArgumentTypeError,
    Namespace,
)
from pathlib import Path
from typing import Sequence

from ._version import __version__
from .formatter.config import DEFAULT_INDENT, Config


class PyProjectFmtNamespace(Namespace):
    """Options for pyproject-fmt tool."""

    inputs: list[Path]
    stdout: bool
    indent: int
    check: bool
    keep_full_version: bool

    @property
    def configs(self) -> list[Config]:
        """:return: configurations"""
        return [
            Config(
                pyproject_toml=toml,
                toml=toml.read_text(encoding="utf-8"),
                indent=self.indent,
                keep_full_version=self.keep_full_version,
            )
            for toml in self.inputs
        ]


def pyproject_toml_path_creator(argument: str) -> Path:
    """
    Validate that pyproject.toml can be formatted.

    :param argument: the string argument passed in
    :return: the pyproject.toml path
    """
    path = Path(argument).absolute()
    if path.is_dir():
        path = path / "pyproject.toml"
    if not path.exists():
        msg = "path does not exist"
        raise ArgumentTypeError(msg)
    if not path.is_file():
        msg = "path is not a file"
        raise ArgumentTypeError(msg)
    if not os.access(path, os.R_OK):
        msg = "cannot read path"
        raise ArgumentTypeError(msg)
    if not os.access(path, os.W_OK):
        msg = "cannot write path"
        raise ArgumentTypeError(msg)
    return path


def _build_cli() -> ArgumentParser:
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        prog="pyproject-fmt",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="print package version of pyproject_fmt",
        version=f"%(prog)s ({__version__})",
    )
    group = parser.add_mutually_exclusive_group()
    msg = "print the formatted text to the stdout (instead of update in-place)"
    group.add_argument("-s", "--stdout", action="store_true", help=msg)
    msg = "check and fail if any input would be formatted, printing any diffs"
    group.add_argument("--check", action="store_true", help=msg)
    msg = "keep full dependency versions. For example do not change version 1.0.0 to 1"
    parser.add_argument("--keep-full-version", action="store_true", help=msg)
    parser.add_argument(
        "--indent",
        type=int,
        default=DEFAULT_INDENT,
        help="number of spaces to indent",
    )
    msg = "pyproject.toml file(s) to format"
    parser.add_argument("inputs", nargs="+", type=pyproject_toml_path_creator, help=msg)
    return parser


def cli_args(args: Sequence[str]) -> PyProjectFmtNamespace:
    """
    Load the tools options.

    :param args: CLI arguments
    :return: the parsed options
    """
    parser = _build_cli()
    result = PyProjectFmtNamespace()
    parser.parse_args(namespace=result, args=args)
    return result


__all__ = [
    "cli_args",
    "PyProjectFmtNamespace",
]



"""Defines configuration for the formatter."""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Final


DEFAULT_INDENT: Final[int] = 2  #: default indentation level


@dataclass(frozen=True)
class Config:
    """Configuration flags for the formatting."""

    pyproject_toml: Path
    toml: str  #: the text to format
    indent: int = DEFAULT_INDENT  #: indentation to apply
    keep_full_version: bool = False


__all__ = [
    "Config",
    "DEFAULT_INDENT",
]
